/*********************************************************************
 * TipMe - full featured tip calculator for the Palm Pilot
 * Alan Harder - v1.3  - 24-Feb-2003
 *               v1.23 - 29-Mar-1999
 *               v1.0  - 20-Oct-1997
 * Alan.Harder@Sun.COM     http://www.moshpit.org/pilot/
 * Freeware - feel free to distribute and/or modify this code.. just
 *            leave my name on it.  This program was developed with gcc
 *            and prc-tools on Solaris 2.6.  Version 1.3 was built with
 *            prc-tools 2.2 and pilrc 2.9 on Solaris 8.
 ********************************************************************/
#include "tipme.h"
#include "floatmgr.h"
#include "GuiTools.h"

#define CREATOR		'TpMe'
#define PREF_VERSION	3

Char deciChr = '.';
Char *amtPat = "%ld%c%00ld", *negPat = "(%ld%c%00ld)";
Int16 divVal[] = { 100, 1000 };

// App Preferences Type
//
typedef struct
{
    Int16 taxRate, tipRate, roundType, tipOn, roundAmt;
    Boolean useTax;
}
TipPrefsType;

// Methods to get and set Amounts in text fields.  The amount is a long
// integer and will be displayed with the decimal point placed appropriately..
//
void PutAmountNInField(FieldPtr fieldPtr, Int32 amount, Int16 decimals)
{
    Int32 amt; Int16 div;
    Char str[16], pat[16];
    if (decimals==3 && amount%10==0) { amount/=10; decimals--; }
    if (amount < 0)
    {
	amt = -amount;
	StrCopy(pat, negPat);
	pat[8] = '0' + decimals;
    } else {
	amt = amount;
	StrCopy(pat, amtPat);
	pat[7] = '0' + decimals;
    }
    div = divVal[decimals-2];
    StrPrintF(str, pat, amt/div, deciChr, amt%div);
    SetText(fieldPtr, str);
}
void PutAmountInField(FieldPtr f, Int32 a) { PutAmountNInField(f, a, 2); }

Int32 GetAmountNFromField(FieldPtr fieldPtr, Int16 decimals)
{
    Int32 val; Int16 div;
    Char *str2, *str = GetText(fieldPtr);
    if (!str) return -1;
    if (!str[0]) { MemPtrFree(str); return -1; }
    val = StrAToI(str) * (div=divVal[decimals-2]);
    str2 = StrChr(str, deciChr);
    while (str2 && str2[1] >= '0' && str2[1] <= '9')
	val += (*++str2 - '0') * (div/=10);
    MemPtrFree(str);
    return val;
}
Int32 GetAmountFromField(FieldPtr f) { GetAmountNFromField(f, 2); }


UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
    EventType evt;
    UInt16 error, prefsSize;
    TipPrefsType *prefs;

    Int16 people, diff, type;
    Int32 total, share, amount, tip, tax, tipbase, roundby;
    FieldPtr amountFld, taxFld, taxAmtFld, tipFld, tipAmtFld,
	     totalFld, peopleFld, shareFld, roundAmtFld, couponFld;
    ControlPtr taxCB, rndUpBtn, rndDownBtn;
    ListPtr rndType, tipOn;

    if (cmd != sysAppLaunchCmdNormalLaunch)
	return 0;

    type = PrefGetPreference(prefNumberFormat);
    if (type==nfPeriodComma || type==nfSpaceComma || type==nfApostropheComma)
	deciChr = ',';

    FrmGotoForm(myForm);

    prefs = MemPtrNew(sizeof(TipPrefsType));
    prefsSize = sizeof(TipPrefsType);
    if (PrefGetAppPreferences(CREATOR, PREF_VERSION,
			      prefs, &prefsSize, true) == noPreferenceFound)
    {
	prefs->tipRate = prefs->taxRate = prefs->roundType = 0;
	prefs->tipOn = 2;
	prefs->useTax = false;
	prefs->roundAmt = 100;
    }

    while (1)
    {
	EvtGetEvent(&evt, evtWaitForever);
	if (SysHandleEvent(&evt))
	    continue;
	if (MenuHandleEvent(NULL, &evt, &error))
	    continue;

	switch (evt.eType)
	{
	  case ctlSelectEvent:
	    switch (evt.data.ctlSelect.controlID)
	    {
	      case calcBtn:
		/* Integer overflow may occur for calculations involving
		 * multiplication.. use UInt32 math if amount <= 4294.96
		 * and tax/tip <= 100%, otherwise use FloatMgr math.
		 * Use nearest rounding for tax, round up for tip/share.
		 */
		amount = GetAmountFromField(amountFld);
		if (amount >= 0)
		    PutAmountInField(amountFld, amount);
		tip = GetAmountFromField(tipFld);
		if (tip >= 0)
		    PutAmountInField(tipFld, tip);
		if (amount < 0 || tip < 0)
		{
		    SndPlaySystemSound(sndWarning);
		    SetFocus( (tip < 0) ? tipField : amountField );
		    break;
		}
		type = LstGetSelection(tipOn);
		tax = GetAmountNFromField(taxFld, 3);
		if (tax >= 0) PutAmountNInField(taxFld, tax, 3);
		else          tax = 0;
		if (CtlGetValue(taxCB) > 0)
		{
		    if (amount <= 85899 && tax <= 50000)
			tax = ((UInt32)amount*tax+50000)/100000;
		    else
			tax = doubleToLong( divDouble( addDouble(
			    mulDouble(longToDouble(amount), longToDouble(tax)),
			    50000.0 ), 100000.0 ) );
		    PutAmountInField(taxAmtFld, tax);
		} else if (type == 1) {  // pre-tax calc
		    if (amount <= 21474 && tax <= 20000)
			tax = ((UInt32)amount*tax*10 / (tax+100000) + 5)/10;
		    else
			tax = doubleToLong( divDouble( addDouble( divDouble(
			    mulDouble(
			    mulDouble(longToDouble(amount), longToDouble(tax)),
			    10.0), addDouble(longToDouble(tax), 100000.0) ),
			    5.0), 10.0) );
		    PutAmountInField(taxAmtFld, tax=-tax);
		} else {
		    tax = 0;
		    SetText(taxAmtFld, "");
		}
		tipbase = GetAmountFromField(couponFld);
		if (tipbase >= 0) PutAmountInField(couponFld, tipbase);
		else              tipbase = 0;
		if (type == 0 || (type == 1 && tax >= 0))
		    tipbase += amount;
		else
		    tipbase += amount + tax;
		if (tipbase <= 429496 && tip <= 10000)
		    tip = ((UInt32)tipbase*tip+9999)/10000;
		else
		    tip = doubleToLong( divDouble( addDouble(
			  mulDouble(longToDouble(tipbase), longToDouble(tip)),
			      9999.0 ), 10000.0 ) );
		PutAmountInField(tipAmtFld, tip);
		if (tax < 0) tax = 0;
		total = amount + tax + tip;
		PutAmountInField(totalFld, total);
		people = GetNumFromField(peopleFld);
		if (people > 0)
		{
		    PutNumInField(peopleFld, people);
		    share = (total + people - 1) / people;
		    PutAmountInField(shareFld, share);
		}
		else
		{
		    share = -1;
		    SetText(shareFld, "");
		    SetText(peopleFld, "");
		}
		// enable rounding buttons:
		CtlSetLabel(rndUpBtn, "\001");
		CtlSetEnabled(rndUpBtn, true);
		CtlSetLabel(rndDownBtn, "\002");
		CtlSetEnabled(rndDownBtn, true);
		FrmDrawForm(FrmGetActiveForm());
		break;

	      case clearBtn:
		for (error = 0; error < 7; error++)
		    SetTextById(amountField + error, "");
		if (prefs->tipRate > 0)
		    PutAmountInField(tipFld, prefs->tipRate);
		if (prefs->taxRate > 0)
		    PutAmountNInField(taxFld, prefs->taxRate, 3);
		CtlSetLabel(rndUpBtn, "\003");
		CtlSetEnabled(rndUpBtn, false);
		CtlSetLabel(rndDownBtn, "\004");
		CtlSetEnabled(rndDownBtn, false);
		SetFocus(amountField);
		FrmDrawForm(FrmGetActiveForm());
		break;

	      case roundUpBtn:
	      case roundDownBtn:
		roundby = GetAmountFromField(roundAmtFld);
		if (roundby <= 0) break;
		type = LstGetSelection(rndType);
		tip = GetAmountFromField(tipAmtFld);
		diff = (evt.data.ctlSelect.controlID==roundUpBtn) ? roundby :0;

		switch (type)
		{
		  case 0: // total
		    diff -= total%roundby;
		    if (diff == 0) diff = -roundby;
		    break;
		  case 1: // tip
		    diff -= tip%roundby;
		    if (diff == 0) diff = -roundby;
		    break;
		  case 2: // share
		    diff -= share%roundby;
		    if (diff == 0) diff = -roundby;
		    diff = (share + diff) * people - total;
		    break;
		}
		if ((total + diff) < 0 || (tip + diff) < 0)
		    break;
		total += diff;
		tip += diff;
		if (people > 0)
		{
		    share = (total + people - 1) / people;
		    PutAmountInField(shareFld, share);
		}
		PutAmountInField(totalFld, total);
		PutAmountInField(tipAmtFld, tip);

		if (tip <= 429496)
		    tip = (UInt32)tip * 10000 / tipbase;
		else
		    tip = doubleToLong( divDouble(
			  mulDouble(longToDouble(tip), 10000.0),
			  longToDouble(tipbase) ) );
		PutAmountInField(tipFld, tip);
		FrmDrawForm(FrmGetActiveForm());
		break;

	      case resetTipBtn:
		if (prefs->tipRate > 0)
		{
		    PutAmountInField(tipFld, prefs->tipRate);
		    FldDrawField(tipFld);
		}
		break;

	      case resetAmtBtn:
		SetText(amountFld, "");
		SetFocus(amountField);
		FldDrawField(amountFld);
		break;

	      default:
		FrmHandleEvent(FrmGetActiveForm(), &evt);
	    }
	    break;

	  case frmLoadEvent:
	    FrmSetActiveForm(FrmInitForm(evt.data.frmLoad.formID));
	    if (evt.data.frmLoad.formID == myForm)
	    {
		FormPtr frm = FrmGetActiveForm();
		amountFld = (FieldPtr)GetObjectPtrFromForm(frm, amountField);
		taxFld = (FieldPtr)GetObjectPtrFromForm(frm, taxField);
		taxAmtFld = (FieldPtr)GetObjectPtrFromForm(frm,taxAmountField);
		tipFld = (FieldPtr)GetObjectPtrFromForm(frm, tipField);
		tipAmtFld = (FieldPtr)GetObjectPtrFromForm(frm,tipAmountField);
		totalFld = (FieldPtr)GetObjectPtrFromForm(frm, totalField);
		peopleFld = (FieldPtr)GetObjectPtrFromForm(frm,peopleField);
		shareFld = (FieldPtr)GetObjectPtrFromForm(frm, shareField);
		couponFld = (FieldPtr)GetObjectPtrFromForm(frm, couponField);
		taxCB = (ControlPtr)GetObjectPtrFromForm(frm, taxCBox);
		rndType = (ListPtr)GetObjectPtrFromForm(frm, roundList);
		rndUpBtn = (ControlPtr)GetObjectPtrFromForm(frm, roundUpBtn);
		rndDownBtn = (ControlPtr)GetObjectPtrFromForm(frm,roundDownBtn);
		roundAmtFld = (FieldPtr)GetObjectPtrFromForm(frm,roundAmtField);
		tipOn = (ListPtr)GetObjectPtrFromForm(frm, tiponList);
		if (prefs->tipRate > 0)
		    PutAmountInField(tipFld, prefs->tipRate);
		if (prefs->taxRate > 0)
		    PutAmountNInField(taxFld, prefs->taxRate, 3);
		if (prefs->roundAmt > 0)
		    PutAmountInField(roundAmtFld, prefs->roundAmt);
		if (prefs->useTax)
		    CtlSetValue(taxCB, 1);
		if (prefs->roundType > 0)
		{
		    LstSetSelection(rndType, prefs->roundType);
		    CtlSetLabel(GetObjectPtrFromForm(frm, roundTrig),
				LstGetSelectionText(rndType,prefs->roundType));
		}
		if (prefs->tipOn > 0)
		{
		    LstSetSelection(tipOn, prefs->tipOn);
		    CtlSetLabel(GetObjectPtrFromForm(frm, tiponTrig),
				LstGetSelectionText(tipOn, prefs->tipOn));
		}
	    }
	    break;

	  case frmOpenEvent:
	    FrmDrawForm(FrmGetActiveForm());
	    if (evt.data.frmOpen.formID == myForm)
		SetFocus( (prefs->tipRate > 0) ? amountField : tipField );
	    break;

	  case menuEvent:
	    if (evt.data.menu.itemID == menuSave)
	    {
		Int32 intVal = GetAmountFromField(tipFld);
		if (intVal >= 0)
		    prefs->tipRate = intVal;
		intVal = GetAmountNFromField(taxFld, 3);
		if (intVal >= 0)
		    prefs->taxRate = intVal;
	    }
	    else
		FrmHandleEvent(FrmGetActiveForm(), &evt);
	    break;

	  case menuCmdBarOpenEvent:
	    MenuCmdBarAddButton(menuCmdBarOnRight, percentBitmap,
		menuCmdBarResultMenuItem, menuSave, 0);
	    evt.data.menuCmdBarOpen.preventFieldButtons = true;
	    FrmHandleEvent(FrmGetActiveForm(), &evt);
	    break;

	  case appStopEvent:
	    {
		Int32 intVal;
		prefs->roundType = LstGetSelection(rndType);
		prefs->tipOn = LstGetSelection(tipOn);
		prefs->useTax = (CtlGetValue(taxCB) != 0);
		intVal = GetAmountFromField(roundAmtFld);
		if (intVal >= 0)
		    prefs->roundAmt = intVal;
		PrefSetAppPreferences(CREATOR, PREF_VERSION, PREF_VERSION,
				      prefs, sizeof(TipPrefsType), true);
		MemPtrFree(prefs);
	    }
	    FrmCloseAllForms();
	    return 0;

	  default:
	    FrmHandleEvent(FrmGetActiveForm(), &evt);
	}
    }
    return 0;
}

